@@ -1,21 +1,22 @@
+import jinja2
+import json
 import logging
+import math
 import mimetypes
 import os
 import pathlib
 from http.server import HTTPServer, BaseHTTPRequestHandler
 from socketserver import ThreadingMixIn
 from datetime import datetime, timedelta
-import jinja2
-import json
 import re
+import subprocess
+import sys
 import threading
 import time
 import urllib.parse
 import uuid
-import math
-from ww import f
 
-logger = logging.getLogger(__name__.rsplit(".")[-1])
+logger = logging.getLogger("\U0001F3AE HTTP")
 
 
 class ThreadingSimpleServer(ThreadingMixIn, HTTPServer):
@@ -66,6 +67,7 @@ def CreateHTTPHandlerClass(master):
     class HTTPControlHandler(BaseHTTPRequestHandler):
         ampsList = []
         fields = {}
+        host = None
         hoursDurationList = []
         master = None
         path = ""
@@ -128,16 +130,21 @@ def CreateHTTPHandlerClass(master):
             # render HTML, we can keep using those even inside jinja2
             self.templateEnv.globals.update(addButton=self.addButton)
             self.templateEnv.globals.update(ampsList=self.ampsList)
+            self.templateEnv.globals.update(
+                apiChallenge=master.getModuleByName("TeslaAPI").getApiChallenge
+            )
             self.templateEnv.globals.update(chargeScheduleDay=self.chargeScheduleDay)
             self.templateEnv.globals.update(checkBox=self.checkBox)
+            self.templateEnv.globals.update(checkForUpdates=master.checkForUpdates)
             self.templateEnv.globals.update(doChargeSchedule=self.do_chargeSchedule)
-            self.templateEnv.globals.update(
-                getMFADevices=master.getModuleByName("TeslaAPI").getMFADevices
-            )
+            self.templateEnv.globals.update(host=self.host)
             self.templateEnv.globals.update(hoursDurationList=self.hoursDurationList)
             self.templateEnv.globals.update(navbarItem=self.navbar_item)
             self.templateEnv.globals.update(optionList=self.optionList)
             self.templateEnv.globals.update(timeList=self.timeList)
+            self.templateEnv.globals.update(
+                vehicles=master.getModuleByName("TeslaAPI").getCarApiVehicles
+            )
 
             # Set master object
             self.master = master
@@ -531,6 +538,22 @@ def CreateHTTPHandlerClass(master):
                 self.send_response(204)
                 self.end_headers()
 
+            elif self.url.path == "/api/sendTeslaAPICommand":
+                data = json.loads(self.post_data.decode("UTF-8"))
+                command = str(data.get("commandName", None))
+                vehicle = str(data.get("vehicleID", None))
+                params = str(data.get("parameters", None))
+
+                res = master.getModuleByName("TeslaAPI").apiDebugInterface(
+                    command, vehicle, params
+                )
+                if res == True:
+                    self.send_response(200)
+                    self.end_headers()
+                else:
+                    self.send_response(400)
+                    self.end_headers()
+
             elif self.url.path == "/api/setSetting":
                 data = json.loads(self.post_data.decode("UTF-8"))
                 setting = str(data.get("setting", None))
@@ -662,6 +685,16 @@ def CreateHTTPHandlerClass(master):
         def do_GET(self):
             self.url = urllib.parse.urlparse(self.path)
 
+            # Determine host header entry
+            try:
+                self.host = self.headers.get("Host", "")
+
+                # Remove port number from domain
+                if ":" in self.host:
+                    self.host = self.host.split(":", 1)[0]
+            except IndexError:
+                self.host = None
+
             # serve local static content files (from './lib/TWCManager/Control/static/' dir)
             if self.url.path.startswith("/static/"):
                 content_type = mimetypes.guess_type(self.url.path)[0]
@@ -697,21 +730,13 @@ def CreateHTTPHandlerClass(master):
                 {"route": "/debug", "tmpl": "debug.html.j2"},
                 {"route": "/schedule", "tmpl": "schedule.html.j2"},
                 {"route": "/settings", "tmpl": "settings.html.j2"},
-                {"route": "/teslaAccount/login", "error": "insecure"},
-                {"route": "/teslaAccount/mfaCode", "error": "insecure"},
-                {"route": "/teslaAccount/submitCaptcha", "error": "insecure"},
+                {"route": "/teslaAccount/saveToken", "error": "insecure"},
                 {"rstart": "/teslaAccount", "tmpl": "main.html.j2"},
+                {"route": "/upgradePrompt", "tmpl": "upgradePrompt.html.j2"},
                 {"rstart": "/vehicleDetail", "tmpl": "vehicleDetail.html.j2"},
                 {"route": "/vehicles", "tmpl": "vehicles.html.j2"},
             ]
 
-            if self.url.path == "/teslaAccount/getCaptchaImage":
-                self.send_response(200)
-                self.send_header("Content-type", "image/svg+xml")
-                self.end_headers()
-                self.wfile.write(master.getModuleByName("TeslaAPI").getCaptchaImage())
-                return
-
             if self.url.path == "/":
                 self.send_response(200)
                 self.send_header("Content-type", "text/html")
@@ -785,6 +810,34 @@ def CreateHTTPHandlerClass(master):
                 self.wfile.write(page.encode("utf-8"))
                 return
 
+            if self.url.path == "/upgrade":
+                # This is extremely beta
+                # Attempt a self-update of TWCManager by calling pip
+                self.send_response(200)
+                self.send_header("Content-type", "text/html")
+                self.end_headers()
+
+                self.template = self.templateEnv.get_template("upgrade.html.j2")
+                page = self.template.render(self.__dict__)
+
+                try:
+                    page += subprocess.check_output(
+                        [
+                            sys.executable,
+                            "-m",
+                            "pip",
+                            "install",
+                            "--user",
+                            "--upgrade",
+                            "TWCManager",
+                        ]
+                    ).decode("UTF-8")
+                except subprocess.CalledProcessError as error:
+                    page += "An error occurred attempting upgrade: " + str(error)
+
+                self.wfile.write(page.encode("utf-8"))
+                return
+
             if self.url.path.startswith("/vehicles/deleteGroup"):
                 group = urllib.parse.unquote(self.url.path.rsplit("/", 1)[1])
                 if (
@@ -855,6 +908,10 @@ def CreateHTTPHandlerClass(master):
                 self.process_save_settings("debug")
                 return
 
+            if self.url.path == "/debug/saveToggle":
+                self.process_save_settings("debug_toggle")
+                return
+
             if self.url.path == "/schedule/save":
                 # User has submitted schedule.
                 self.process_save_schedule()
@@ -866,38 +923,6 @@ def CreateHTTPHandlerClass(master):
                 self.process_save_settings()
                 return
 
-            if self.url.path == "/teslaAccount/login":
-                # User has submitted Tesla login.
-                # Pass it to the dedicated process_teslalogin function
-                self.process_teslalogin()
-                return
-
-            if self.url.path == "/teslaAccount/mfaCode":
-                transactionID = self.getFieldValue("transactionID")
-                mfaDevice = self.getFieldValue("mfaDevice")
-                mfaCode = self.getFieldValue("mfaCode")
-
-                resp = master.getModuleByName("TeslaAPI").mfaLogin(
-                    transactionID, mfaDevice, mfaCode
-                )
-
-                self.send_response(302)
-                self.send_header("Location", "/teslaAccount/" + str(resp))
-                self.end_headers()
-                self.wfile.write("".encode("utf-8"))
-                return
-
-            if self.url.path == "/teslaAccount/submitCaptcha":
-                captchaCode = self.getFieldValue("captchaCode")
-
-                resp = master.getModuleByName("TeslaAPI").submitCaptchaCode(captchaCode)
-
-                self.send_response(302)
-                self.send_header("Location", "/teslaAccount/" + str(resp))
-                self.end_headers()
-                self.wfile.write("".encode("utf-8"))
-                return
-
             if self.url.path == "/graphs/dates":
                 # User has submitted dates to graph this period.
                 objIni = self.getFieldValue("dateIni")
@@ -918,6 +943,33 @@ def CreateHTTPHandlerClass(master):
                 self.wfile.write("".encode("utf-8"))
                 return
 
+            if self.url.path == "/teslaAccount/saveToken":
+
+                # Check if we are skipping Tesla Login submission
+                later = False
+                try:
+                    later = len(self.fields["later"][0])
+                except KeyError:
+                    later = False
+
+                res = ""
+                url = self.getFieldValue("url")
+
+                if later:
+                    master.teslaLoginAskLater = True
+                    res = "later"
+
+                else:
+
+                    res = master.getModuleByName("TeslaAPI").saveApiToken(url)
+
+                self.send_response(302)
+                self.send_header("Location", "/teslaAccount/" + res)
+
+                self.end_headers()
+                self.wfile.write("".encode("utf-8"))
+                return
+
             if self.url.path == "/vehicle/groupMgmt":
 
                 group = self.getFieldValue("group")
@@ -1175,25 +1227,32 @@ def CreateHTTPHandlerClass(master):
                     carapi = master.getModuleByName("TeslaAPI")
                     if key == "carApiBearerToken":
                         carapi.setCarApiBearerToken(self.getFieldValue(key))
+                        # New tokens expire after 8 hours
+                        carapi.setCarApiTokenExpireTime(time.time() + 8 * 60 * 60)
                     elif key == "carApiRefreshToken":
                         carapi.setCarApiRefreshToken(self.getFieldValue(key))
+                        carapi.setCarApiTokenExpireTime(time.time() + 45 * 24 * 60 * 60)
 
-                # Write setting to dictionary
-                master.settings[key] = self.getFieldValue(key)
+                else:
+
+                    # Write setting to dictionary
+                    master.settings[key] = self.getFieldValue(key)
 
             # If Non-Scheduled power action is either Do not Charge or
             # Track Green Energy, set Non-Scheduled power rate to 0
             if int(master.settings.get("nonScheduledAction", 1)) > 1:
                 master.settings["nonScheduledAmpsMax"] = 0
-            master.queue_background_task({"cmd": "saveSettings"})
 
             # If triggered from the Debug page (not settings page), we need to
             # set certain settings to false if they were not seen in the
             # request data - This is because Check Boxes don't have a value
             # if they aren't set
+            if page == "debug_toggle":
+                if "enableDebugCommands" not in self.fields:
+                    master.settings["enableDebugCommands"] = 0
+
             if page == "debug":
                 checkboxes = [
-                    "enableDebugCommands",
                     "spikeAmpsProactively",
                     "spikeAmpsReactively",
                 ]
@@ -1201,6 +1260,9 @@ def CreateHTTPHandlerClass(master):
                     if checkbox not in self.fields:
                         master.settings[checkbox] = 0
 
+            # Save Settings
+            master.queue_background_task({"cmd": "saveSettings"})
+
             # Redirect to the index page
             self.send_response(302)
             self.send_header("Location", "/")
@@ -1208,51 +1270,6 @@ def CreateHTTPHandlerClass(master):
             self.wfile.write("".encode("utf-8"))
             return
 
-        def process_teslalogin(self):
-            # Check if we are skipping Tesla Login submission
-
-            if not master.teslaLoginAskLater:
-                later = False
-                try:
-                    later = len(self.fields["later"][0])
-                except KeyError:
-                    later = False
-
-                if later:
-                    master.teslaLoginAskLater = True
-
-            if not master.teslaLoginAskLater:
-                # Connect to Tesla API
-
-                carapi = master.getModuleByName("TeslaAPI")
-                carapi.resetCarApiLastErrorTime()
-                try:
-                    ret = carapi.apiLogin(
-                        self.fields["email"][0], self.fields["password"][0]
-                    )
-                except KeyError:
-                    self.send_response(302)
-                    self.send_header("Location", "/teslaAccount/NotSpecified")
-                    self.end_headers()
-                    self.wfile.write("".encode("utf-8"))
-                    return
-
-                # Redirect to an index page with output based on the return state of
-                # the function
-                self.send_response(302)
-                self.send_header("Location", "/teslaAccount/" + str(ret))
-                self.end_headers()
-                self.wfile.write("".encode("utf-8"))
-                return
-            else:
-                # User has asked to skip Tesla Account submission for this session
-                # Redirect back to /
-                self.send_response(302)
-                self.send_header("Location", "/")
-                self.end_headers()
-                self.wfile.write("".encode("utf-8"))
-                return
-
         def process_save_graphs(self, initial, end):
             # Check that Graphs dict exists within settings.
             # If not, this would indicate that this is the first time
