@@ -27,9 +27,7 @@
 #
 # For more information, please visit http://unlicense.org
 
-import commentjson
 import importlib
-import json
 import logging
 import os.path
 import math
@@ -37,9 +35,9 @@ import re
 import sys
 import time
 import traceback
-from datetime import datetime
+import datetime
+import yaml
 import threading
-from ww import f
 from TWCManager.TWCMaster import TWCMaster
 import requests
 from enum import Enum
@@ -65,7 +63,7 @@ logging.INFO9 = 12
 logging.DEBUG2 = 9
 
 
-logger = logging.getLogger("TWCManager")
+logger = logging.getLogger("\u26FD Manager")
 
 # Define available modules for the instantiator
 # All listed modules will be loaded at boot time
@@ -83,11 +81,13 @@ modules_available = [
     "Interface.TCP",
     "Policy.Policy",
     "Vehicle.TeslaAPI",
+    "Vehicle.TeslaMateVehicle",
     "Control.WebIPCControl",
     "Control.HTTPControl",
     "Control.MQTTControl",
     #    "Control.OCPPControl",
     "EMS.Efergy",
+    "EMS.EmonCMS",
     "EMS.Enphase",
     "EMS.Fronius",
     "EMS.Growatt",
@@ -104,6 +104,7 @@ modules_available = [
     "EMS.TeslaPowerwall2",
     "EMS.TED",
     "EMS.Volkszahler",
+    "EMS.URL",
     "Status.HASSStatus",
     "Status.MQTTStatus",
 ]
@@ -126,7 +127,15 @@ else:
         jsonconfig = open("config.json")
 
 if jsonconfig:
-    config = commentjson.load(jsonconfig)
+    configtext = ""
+    for line in jsonconfig:
+        if line.lstrip().startswith("//") or line.lstrip().startswith("#"):
+            configtext += "\n"
+        else:
+            configtext += line.replace("\t", " ").split("#")[0]
+
+    config = yaml.safe_load(configtext)
+    configtext = None
 else:
     logger.error("Unable to find a configuration file.")
     sys.exit()
@@ -186,7 +195,7 @@ def hex_str(ba: bytearray):
 
 def time_now():
     global config
-    return datetime.now().strftime(
+    return datetime.datetime.now().strftime(
         "%H:%M:%S" + (".%f" if config["config"]["displayMilliseconds"] else "")
     )
 
@@ -309,6 +318,8 @@ def background_tasks_thread(master):
                         requests.post(task["url"], json=body)
                 elif task["cmd"] == "saveSettings":
                     master.saveSettings()
+                elif task["cmd"] == "sunrise":
+                    update_sunrise_sunset()
 
         except:
             logger.info(
@@ -352,7 +363,7 @@ def update_statuses():
     # Print a status update if we are on track green energy showing the
     # generation and consumption figures
     maxamps = master.getMaxAmpsToDivideAmongSlaves()
-    maxampsDisplay = f("{maxamps:.2f}A")
+    maxampsDisplay = f"{maxamps:.2f}A"
     if master.getModuleByName("Policy").policyIsGreen():
         genwatts = master.getGeneration()
         conwatts = master.getConsumption()
@@ -380,9 +391,9 @@ def update_statuses():
 
             logger.info(
                 "Green energy Generates %s, Consumption %s (Charger Load %s)",
-                f("{genwatts:.0f}W"),
-                f("{conwatts:.0f}W"),
-                f("{chgwatts:.0f}W"),
+                f"{genwatts:.0f}W",
+                f"{conwatts:.0f}W",
+                f"{chgwatts:.0f}W",
                 extra=logExtra,
             )
 
@@ -390,10 +401,10 @@ def update_statuses():
 
             logger.info(
                 "Green energy Generates %s, Consumption %s (Charger Load %s, Other Load %s)",
-                f("{genwatts:.0f}W"),
-                f("{conwatts:.0f}W"),
-                f("{chgwatts:.0f}W"),
-                f("{othwatts:.0f}W"),
+                f"{genwatts:.0f}W",
+                f"{conwatts:.0f}W",
+                f"{chgwatts:.0f}W",
+                f"{othwatts:.0f}W",
                 extra=logExtra,
             )
 
@@ -401,11 +412,11 @@ def update_statuses():
 
             logger.info(
                 "Green energy Generates %s, Consumption %s (Charger Load %s, Other Load %s, Offset %s)",
-                f("{genwatts:.0f}W"),
-                f("{conwatts:.0f}W"),
-                f("{chgwatts:.0f}W"),
-                f("{othwatts:.0f}W"),
-                f("{conoffset:.0f}W"),
+                f"{genwatts:.0f}W",
+                f"{conwatts:.0f}W",
+                f"{chgwatts:.0f}W",
+                f"{othwatts:.0f}W",
+                f"{conoffset:.0f}W",
                 extra=logExtra,
             )
 
@@ -413,11 +424,11 @@ def update_statuses():
 
             logger.info(
                 "Green energy Generates %s (Offset %s), Consumption %s (Charger Load %s, Other Load %s)",
-                f("{genwatts:.0f}W"),
-                f("{(-1 * conoffset):.0f}W"),
-                f("{conwatts:.0f}W"),
-                f("{chgwatts:.0f}W"),
-                f("{othwatts:.0f}W"),
+                f"{genwatts:.0f}W",
+                f"{(-1 * conoffset):.0f}W",
+                f"{conwatts:.0f}W",
+                f"{chgwatts:.0f}W",
+                f"{othwatts:.0f}W",
                 extra=logExtra,
             )
 
@@ -438,15 +449,13 @@ def update_statuses():
             )
         )
         if abs(maxamps - nominalOffer) > 0.005:
-            nominalOfferDisplay = f("{nominalOffer:.2f}A")
+            nominalOfferDisplay = f"{nominalOffer:.2f}A"
             logger.debug(
-                f(
-                    "Offering {maxampsDisplay} instead of {nominalOfferDisplay} to compensate for inexact current draw"
-                )
+                f"Offering {maxampsDisplay} instead of {nominalOfferDisplay} to compensate for inexact current draw"
             )
             conwatts = genwatts - master.convertAmpsToWatts(maxamps)
-        generation = f("{master.convertWattsToAmps(genwatts):.2f}A")
-        consumption = f("{master.convertWattsToAmps(conwatts):.2f}A")
+        generation = f"{master.convertWattsToAmps(genwatts):.2f}A"
+        consumption = f"{master.convertWattsToAmps(conwatts):.2f}A"
         logger.info(
             "Limiting charging to %s - %s = %s.",
             generation,
@@ -462,7 +471,7 @@ def update_statuses():
         )
 
     # Print minimum charge for all charging policies
-    minchg = f("{config['config']['minAmpsPerTWC']}A")
+    minchg = f"{config['config']['minAmpsPerTWC']}A"
     logger.info(
         "Charge when above %s (minAmpsPerTWC).", minchg, extra={"colored": "magenta"}
     )
@@ -485,6 +494,60 @@ def update_statuses():
         )
 
 
+def update_sunrise_sunset():
+
+    ltNow = time.localtime()
+    latlong = master.getHomeLatLon()
+    if latlong[0] == 10000:
+        # We don't know where home is; keep defaults
+        master.settings["sunrise"] = 6
+        master.settings["sunset"] = 20
+    else:
+        sunrise = 6
+        sunset = 20
+        url = (
+            "https://api.sunrise-sunset.org/json?lat="
+            + str(latlong[0])
+            + "&lng="
+            + str(latlong[1])
+            + "&formatted=0&date="
+            + "-".join([str(ltNow.tm_year), str(ltNow.tm_mon), str(ltNow.tm_mday)])
+        )
+
+        r = {}
+        try:
+            r = requests.get(url).json().get("results")
+        except:
+            pass
+
+        if r.get("sunrise", None):
+            try:
+                dtSunrise = datetime.datetime.astimezone(
+                    datetime.datetime.fromisoformat(r["sunrise"])
+                )
+                sunrise = dtSunrise.hour + (1 if dtSunrise.minute >= 30 else 0)
+            except:
+                pass
+
+        if r.get("sunset", None):
+            try:
+                dtSunset = datetime.datetime.astimezone(
+                    datetime.datetime.fromisoformat(r["sunset"])
+                )
+                sunset = dtSunset.hour + (1 if dtSunset.minute >= 30 else 0)
+            except:
+                pass
+
+        master.settings["sunrise"] = sunrise
+        master.settings["sunset"] = sunset
+
+    tomorrow = datetime.datetime.combine(
+        datetime.datetime.today(), datetime.time(hour=1)
+    ) + datetime.timedelta(days=1)
+    diff = tomorrow - datetime.datetime.now()
+    master.queue_background_task({"cmd": "sunrise"}, diff.total_seconds())
+
+
 #
 # End functions
 #
@@ -586,6 +649,8 @@ backgroundTasksThread = threading.Thread(target=background_tasks_thread, args=(m
 backgroundTasksThread.daemon = True
 backgroundTasksThread.start()
 
+master.queue_background_task({"cmd": "sunrise"}, 30)
+
 logger.info(
     "TWC Manager starting as fake %s with id %02X%02X and sign %02X"
     % (
