@@ -9,12 +9,12 @@ import queue
 from sys import modules
 import threading
 import time
-from ww import f
 import math
 import random
+import requests
 import bisect
 
-logger = logging.getLogger(__name__.rsplit(".")[-1])
+logger = logging.getLogger("\u26FD Master")
 
 
 class TWCMaster:
@@ -32,6 +32,7 @@ class TWCMaster:
     lastkWhPoll = 0
     lastSaveFailed = 0
     lastTWCResponseMsg = None
+    lastUpdateCheck = 0
     masterTWCID = ""
     maxAmpsToDivideAmongSlaves = 0
     modules = {}
@@ -65,7 +66,8 @@ class TWCMaster:
     subtractChargerLoad = False
     teslaLoginAskLater = False
     TWCID = None
-    version = "1.2.3"
+    updateVersion = False
+    version = "1.2.6"
 
     # TWCs send a seemingly-random byte after their 2-byte TWC id in a number of
     # messages. I call this byte their "Sign" for lack of a better term. The byte
@@ -105,6 +107,50 @@ class TWCMaster:
     def cancelStopCarsCharging(self):
         self.delete_background_task({"cmd": "charge", "charge": False})
 
+    def checkForUpdates(self):
+        # This function is used by the Web UI and later on will be used by the console to detect TWCManager Updates
+        # It runs a maximum of once per hour, and queries the current PyPi package version vs the current version
+        # If they match, it returns false. If there's a different version available, we alert the user
+        if time.time() > self.lastUpdateCheck + (60 * 60):
+            self.lastUpdateCheck = time.time()
+
+            # Fetch the JSON data from PyPi for our package
+            url = "https://pypi.org/pypi/twcmanager/json"
+
+            try:
+                req = requests.get(url)
+                logger.log(logging.INFO8, "Requesting PyPi package info " + str(req))
+                pkgInfo = json.loads(req.text)
+            except requests.exceptions.RequestException:
+                logger.info("Failed to fetch package details " + url)
+                logger.log(logging.INFO6, "Response: " + req.text)
+                pass
+            except json.decoder.JSONDecodeError:
+                logger.info("Could not parse JSON result from " + url)
+                logger.log(logging.INFO6, "Response: " + req.text)
+                pass
+
+            if pkgInfo.get("info", {}).get("version", None):
+                if pkgInfo["info"]["version"] != self.version:
+                    # Versions don't match. Let's make sure the new one really is newer
+                    current_arr = [int(v) for v in self.version.split(".")]
+                    avail_arr = [int(v) for v in pkgInfo["info"]["version"].split(".")]
+                    for i in range(max(len(current_arr), len(avail_arr))):
+                        v1 = current_arr[i] if i < len(current_arr) else 0
+                        v2 = avail_arr[i] if i < len(avail_arr) else 0
+
+                        # If any element of current version in order from first to last is lower than available version,
+                        # advertise newer version
+                        if v1 < v2:
+                            self.updateVersion = pkgInfo["info"]["version"]
+                            break
+
+                        # If current version is greater than available version, do not advertise newer version
+                        if v1 > v2:
+                            break
+
+        return self.updateVersion
+
     def checkModuleCapability(self, type, capability):
         # For modules which advertise capabilities, scan all loaded modules of a certain type and
         # report on if any of those modules advertise the reported capability
@@ -729,6 +775,11 @@ class TWCMaster:
                 "Built-in": 1,
                 "Members": [],
             }
+        # Fill in old defaults as bridge
+        if not self.settings.get("sunrise", None):
+            self.settings["sunrise"] = 6
+        if not self.settings.get("sunset", None):
+            self.settings["sunset"] = 20
 
     def master_id_conflict(self):
         # We're playing fake slave, and we got a message from a master with our TWCID.
@@ -1369,6 +1420,18 @@ class TWCMaster:
             "%H:%M:%S" + (".%f" if self.config["config"]["displayMilliseconds"] else "")
         )
 
+    def tokenSyncEnabled(self):
+        # TODO: Should not be hardcoded
+        # Check if any modules are performing token sync from other projects or interfaces
+        # if so, we do not prompt locally for authentication and we don't use our own settings
+        tokenSync = False
+
+        if self.getModuleByName("TeslaMateVehicle"):
+            if self.getModuleByName("TeslaMateVehicle").syncTokens:
+                tokenSync = True
+
+        return tokenSync
+
     def translateModuleNameToConfig(self, modulename):
         # This function takes a module name (eg. EMS.Fronius) and returns a config section (Sources.Fronius)
         # It makes it easier for us to determine where a module's config should be
