@@ -5,20 +5,20 @@ import logging
 import os
 import re
 import requests
+from threading import Thread
 import time
 from urllib.parse import parse_qs
-from ww import f
 
-logger = logging.getLogger(__name__.rsplit(".")[-1])
+logger = logging.getLogger("\U0001F697 TeslaAPI")
 
 
 class TeslaAPI:
 
-    __apiCaptcha = None
-    __apiCaptchaCode = None
-    authURL = "https://auth.tesla.com/oauth2/v3/authorize"
-    callbackURL = "https://auth.tesla.com/void/callback"
-    captchaURL = "https://auth.tesla.com/captcha"
+    __apiChallenge = None
+    __apiVerifier = None
+    __apiState = None
+    __authURL = "https://auth.tesla.com/oauth2/v3/token"
+    __callbackURL = "https://auth.tesla.com/void/callback"
     carApiLastErrorTime = 0
     carApiBearerToken = ""
     carApiRefreshToken = ""
@@ -39,10 +39,9 @@ class TeslaAPI:
     minChargeLevel = -1
     params = None
     __password = None
-    refreshURL = "https://owner-api.teslamotors.com/oauth/token"
+    refreshURL = "https://auth.tesla.com/oauth2/v3/token"
     __resp = None
     session = None
-    verifier = ""
 
     # Transient errors are ones that usually disappear if we retry the car API
     # command a minute or less later.
@@ -59,9 +58,6 @@ class TeslaAPI:
         "vehicle unavailable",
     ]
 
-    # Define minutes between retrying non-transient errors.
-    carApiErrorRetryMins = 10
-
     def __init__(self, master):
         self.master = master
         try:
@@ -73,193 +69,47 @@ class TeslaAPI:
         except KeyError:
             pass
 
+        self.generateChallenge()
+
     def addVehicle(self, json):
         self.carApiVehicles.append(CarApiVehicle(json, self, self.config))
         return True
 
-    def apiLogin(self, email, password):
+    def apiDebugInterface(self, command, vehicleID, parameters):
 
-        # Populate auth details for Phase 1
-        self.__email = email
-        self.__password = password
+        # Provides an interface from the Web UI to allow commands to be run interactively
 
-        for attempt in range(self.maxLoginRetries):
-
-            self.verifier = base64.urlsafe_b64encode(os.urandom(86)).rstrip(b"=")
-            challenge = base64.urlsafe_b64encode(
-                hashlib.sha256(self.verifier).digest()
-            ).rstrip(b"=")
-            state = (
-                base64.urlsafe_b64encode(os.urandom(16)).rstrip(b"=").decode("utf-8")
-            )
-
-            self.params = (
-                ("client_id", "ownerapi"),
-                ("code_challenge", challenge),
-                ("code_challenge_method", "S256"),
-                ("redirect_uri", self.callbackURL),
-                ("response_type", "code"),
-                ("scope", "openid email offline_access"),
-                ("state", state),
-            )
-
-            self.session = requests.Session()
-            self.__resp = self.session.get(self.authURL, params=self.params)
-
-            if self.__resp.ok and "<title>" in self.__resp.text:
-                logger.log(
-                    logging.INFO6,
-                    "Tesla Auth form fetch success, attempt: " + str(attempt),
-                )
-
-                if 'img data-id="captcha"' in self.__resp.text:
-                    logger.log(
-                        logging.INFO6,
-                        "Tesla Auth form challenged us for Captcha. Redirecting.",
-                    )
-                    self.getApiCaptcha()
-                    return "Phase1Captcha"
-                else:
-                    return self.apiLoginPhaseOne()
-            else:
-                logger.log(
-                    logging.INFO6,
-                    "Tesla auth form fetch failed, attempt: " + str(attempt),
-                )
-
-            time.sleep(3)
-        else:
-            logger.log(
-                logging.INFO2,
-                "Wasn't able to find authentication form after "
-                + str(attempt)
-                + " attempts",
-            )
-            return "Phase1Error"
-
-    def apiLoginPhaseOne(self):
-
-        # Picks up on the first phase of authentication, after redirecting to
-        # handle Captcha if this was requested, or directly if we were lucky
-        # enough not to be challenged.
-
-        csrf = re.search(r'name="_csrf".+value="([^"]+)"', self.__resp.text).group(1)
-        transaction_id = re.search(
-            r'name="transaction_id".+value="([^"]+)"', self.__resp.text
-        ).group(1)
-
-        if not csrf or not transaction_id:
-            # These two parameters are required for Phase 1 (Authentication) auth
-            # If they are missing, we raise an appropriate error to the user's attention
-            return "Phase1Error"
-
-        data = {
-            "_csrf": csrf,
-            "_phase": "authenticate",
-            "_process": "1",
-            "transaction_id": transaction_id,
-            "cancel": "",
-            "identity": self.__email,
-            "credential": self.__password,
-        }
-
-        # If a captcha code is stored, inject it into the data parameter
-        if self.__apiCaptchaCode:
-            data["captcha"] = self.__apiCaptchaCode
-
-            # Clear captcha data
-            self.__apiCaptcha = None
-
-        # Clear stored credentials
-        self.__email = None
-        self.__password = None
-
-        # Call login Phase 2
-        return self.apiLoginPhaseTwo(data)
-
-    def apiLoginPhaseTwo(self, data):
-
-        for attempt in range(self.maxLoginRetries):
-            resp = self.session.post(
-                self.authURL, params=self.params, data=data, allow_redirects=False
-            )
-            if resp.ok and (resp.status_code == 302 or "<title>" in resp.text):
-                logger.log(
-                    logging.INFO2,
-                    "Posted auth form successfully after " + str(attempt) + " attempts",
-                )
-                break
-            time.sleep(3)
-        else:
-            logger.log(
-                logging.INFO2,
-                "Wasn't able to post authentication form after "
-                + str(attempt)
-                + " attempts",
-            )
-            return "Phase2Error"
-
-        if resp.status_code == 200 and "/mfa/verify" in resp.text:
-            # This account is using MFA, redirect to MFA code entry page
-            return "MFA/" + str(data["transaction_id"])
+        # Map vehicle ID back to vehicle object
+        vehicle = self.getVehicleByID(int(vehicleID))
 
+        # Get parameters
+        params = {}
         try:
-            code = parse_qs(resp.headers["location"])[self.callbackURL + "?code"]
-        except KeyError:
-            return "Phase2ErrorTip"
+            params = json.loads(parameters)
+        except json.decoder.JSONDecodeError:
+            pass
 
-        data = {
-            "grant_type": "authorization_code",
-            "client_id": "ownerapi",
-            "code_verifier": self.verifier.decode("utf-8"),
-            "code": code,
-            "redirect_uri": self.callbackURL,
-        }
-
-        resp = self.session.post("https://auth.tesla.com/oauth2/v3/token", json=data)
-        access_token = resp.json()["access_token"]
-
-        headers = {"authorization": "bearer " + access_token}
-
-        data = {
-            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
-            "client_id": self.clientID,
-        }
-        resp = self.session.post(
-            "https://owner-api.teslamotors.com/oauth/token", headers=headers, json=data
-        )
-        try:
-            self.setCarApiBearerToken(resp.json()["access_token"])
-            self.setCarApiRefreshToken(resp.json()["refresh_token"])
-            self.setCarApiTokenExpireTime(time.time() + resp.json()["expires_in"])
-            self.master.queue_background_task({"cmd": "saveSettings"})
+        # Execute specified command
+        if command == "setChargeRate":
+            charge_rate = params.get("charge_rate", 0)
+            self.setChargeRate(charge_rate, vehicle)
+            return True
+        elif command == "wakeVehicle":
+            self.wakeVehicle(vehicle)
             return True
 
-        except KeyError:
-            logger.log(
-                logging.INFO2,
-                "ERROR: Can't access Tesla car via API.  Please log in again via web interface.",
-            )
-            self.updateCarApiLastErrorTime()
-            # In addition to setting carApiLastErrorTime, erase tokens to
-            # prevent further authorization attempts until user enters password
-            # on web interface. I feel this is safer than trying to log in every
-            # ten minutes with a bad token because Tesla might decide to block
-            # remote access to your car after too many authorization errors.
-            self.setCarApiBearerToken("")
-            self.setCarApiRefreshToken("")
-            self.master.queue_background_task({"cmd": "saveSettings"})
-            return False
+        # If we make it here, we did not execute a command
+        return False
 
     def apiRefresh(self):
         # Refresh tokens expire in 45
         # days when first issued, so we'll get a new token every 15 days.
         headers = {"accept": "application/json", "Content-Type": "application/json"}
         data = {
-            "client_id": self.clientID,
-            "client_secret": self.clientSecret,
+            "client_id": "ownerapi",
             "grant_type": "refresh_token",
             "refresh_token": self.getCarApiRefreshToken(),
+            "scope": "openid email offline_access",
         }
         req = None
         now = time.time()
@@ -268,10 +118,16 @@ class TeslaAPI:
             logger.log(logging.INFO2, "Car API request" + str(req))
             apiResponseDict = json.loads(req.text)
         except requests.exceptions.RequestException:
+            logger.log(
+                logging.INFO2, "Request Exception parsing API Token Refresh Response"
+            )
             pass
         except ValueError:
             pass
         except json.decoder.JSONDecodeError:
+            logger.log(
+                logging.INFO2, "JSON Decode Error parsing API Token Refresh Response"
+            )
             pass
 
         try:
@@ -296,6 +152,8 @@ class TeslaAPI:
             self.setCarApiBearerToken("")
             self.setCarApiRefreshToken("")
             self.master.queue_background_task({"cmd": "saveSettings"})
+        except UnboundLocalError:
+            pass
 
     def car_api_available(
         self, email=None, password=None, charge=None, applyLimit=None
@@ -330,9 +188,9 @@ class TeslaAPI:
             )
 
         # Authentiate to Tesla API
-        if (
+        if not self.master.tokenSyncEnabled() and (
             self.getCarApiBearerToken() == ""
-            or self.getCarApiTokenExpireTime() - now < 30 * 24 * 60 * 60
+            or self.getCarApiTokenExpireTime() - now < 60 * 60
         ):
             if self.getCarApiRefreshToken() != "":
                 headers = {
@@ -452,22 +310,7 @@ class TeslaAPI:
 
                     # It's been delayNextWakeAttempt seconds since we last failed to
                     # wake the car, or it's never been woken. Wake it.
-                    vehicle.lastAPIAccessTime = now
-                    url = "https://owner-api.teslamotors.com/api/1/vehicles/"
-                    url = url + str(vehicle.ID) + "/wake_up"
-
-                    headers = {
-                        "accept": "application/json",
-                        "Authorization": "Bearer " + self.getCarApiBearerToken(),
-                    }
-                    try:
-                        req = requests.post(url, headers=headers)
-                        logger.log(logging.INFO8, "Car API cmd wake_up" + str(req))
-                        apiResponseDict = json.loads(req.text)
-                    except requests.exceptions.RequestException:
-                        pass
-                    except json.decoder.JSONDecodeError:
-                        pass
+                    apiResponseDict = self.wakeVehicle(vehicle)
 
                     state = "error"
                     logger.debug("Car API wake car response" + str(apiResponseDict))
@@ -654,6 +497,22 @@ class TeslaAPI:
 
         return True
 
+    def generateChallenge(self):
+        self.__apiVerifier = base64.urlsafe_b64encode(os.urandom(86)).rstrip(b"=")
+        self.__apiChallenge = base64.urlsafe_b64encode(
+            hashlib.sha256(self.__apiVerifier).digest()
+        ).rstrip(b"=")
+        self.__apiState = (
+            base64.urlsafe_b64encode(os.urandom(16)).rstrip(b"=").decode("utf-8")
+        )
+
+    def getApiChallenge(self):
+        return (
+            self.__apiChallenge.decode("UTF-8"),
+            self.__apiState,
+            self.__apiVerifier,
+        )
+
     def is_location_home(self, lat, lon):
 
         if self.master.getHomeLatLon()[0] == 10000:
@@ -668,6 +527,7 @@ class TeslaAPI:
             self.master.setHomeLat(lat)
             self.master.setHomeLon(lon)
             self.master.queue_background_task({"cmd": "saveSettings"})
+            self.master.queue_background_task({"cmd": "sunrise"})
             return True
 
         # 1 lat or lon = ~364488.888 feet. The exact feet is different depending
@@ -1078,24 +938,6 @@ class TeslaAPI:
         if checkArrival:
             self.updateChargeAtHome()
 
-    def getApiCaptcha(self):
-        # This will fetch the current Captcha image displayed by Tesla's auth
-        # website, and store it in memory
-
-        self.__apiCaptcha = self.session.get(self.captchaURL)
-
-    def getCaptchaImage(self):
-        # This will serve the Tesla Captcha image
-
-        if self.__apiCaptcha:
-            return self.__apiCaptcha.content
-        else:
-            logger.log(
-                logging.INFO2,
-                "ERROR: Captcha image requested, but we have none buffered. This is likely due to a stale login session, but if you see it regularly, please report it.",
-            )
-            return ""
-
     def getCarApiBearerToken(self):
         return self.carApiBearerToken
 
@@ -1103,7 +945,8 @@ class TeslaAPI:
         errorCount = self.errorCount
         if vehicle:
             errorCount = max(vehicle.errorCount, errorCount)
-        return pow(2, min(max(errorCount - 1, 0), 6))
+        errorCount = max(errorCount - 1, 0)
+        return min(errorCount, 10)
 
     def getCarApiLastErrorTime(self):
         return self.carApiLastErrorTime
@@ -1147,6 +990,13 @@ class TeslaAPI:
     def getLastStartOrStopChargeTime(self):
         return int(self.carApiLastStartOrStopChargeTime)
 
+    def getVehicleByID(self, vehicleID):
+        # Returns the vehicle object identified by the given ID
+        for vehicle in self.getCarApiVehicles():
+            if vehicle.ID == vehicleID:
+                return vehicle
+        return False
+
     def getVehicleCount(self):
         # Returns the number of currently tracked vehicles
         return int(len(self.carApiVehicles))
@@ -1154,66 +1004,6 @@ class TeslaAPI:
     def getCarApiVehicles(self):
         return self.carApiVehicles
 
-    def getMFADevices(self, transaction_id):
-        # Requests a list of devices we can use for MFA
-        url = f(
-            "https://auth.tesla.com/oauth2/v3/authorize/mfa/factors?transaction_id={transaction_id}"
-        )
-        resp = self.session.get(url)
-        try:
-            content = json.loads(resp.text)
-        except ValueError:
-            return False
-        except json.decoder.JSONDecodeError:
-            return False
-
-        if resp.status_code == 200:
-            return content["data"]
-        elif resp.status_code == 400:
-            logger.error(
-                "The following error was returned when attempting to fetch MFA devices for Tesla Login:"
-                + str(content.get("error", ""))
-            )
-        else:
-            logger.error(
-                "An unexpected error code ("
-                + str(resp.status)
-                + ") was returned when attempting to fetch MFA devices for Tesla Login"
-            )
-
-    def mfaLogin(self, transactionID, mfaDevice, mfaCode):
-        data = {
-            "transaction_id": transactionID,
-            "factor_id": mfaDevice,
-            "passcode": str(mfaCode).rjust(6, "0"),
-        }
-        url = "https://auth.tesla.com/oauth2/v3/authorize/mfa/verify"
-        resp = self.session.post(url, json=data)
-
-        try:
-            jsonData = json.loads(resp.text)
-        except ValueError:
-            return False
-        except json.decoder.JSONDecodeError:
-            return False
-
-        if (
-            "error" in resp.text
-            or not jsonData.get("data", None)
-            or not jsonData["data"].get("approved", None)
-            or not jsonData["data"].get("valid", None)
-        ):
-            if (
-                jsonData.get("error", {}).get("message", None)
-                == "Invalid Attributes: Your passcode should be six digits."
-            ):
-                return "TokenLengthError"
-            else:
-                return "TokenFail"
-        else:
-            data = {"transaction_id": transactionID}
-            return self.apiLoginPhaseTwo(data)
-
     def resetCarApiLastErrorTime(self, vehicle=None):
         self.carApiLastErrorTime = 0
         if vehicle:
@@ -1222,10 +1012,73 @@ class TeslaAPI:
         self.errorCount = 0
         return True
 
+    def saveApiToken(self, url):
+        # Extract code from url
+        if isinstance(url, bytes):
+            url = url.decode("UTF-8")
+        code = re.search(r"code=(.+)&state=(.+)", url)
+
+        logger.log(logging.INFO2, "Code: " + code.group(1))
+        logger.log(logging.INFO2, "State: " + code.group(2))
+
+        # Exchange auth code for bearer token
+        headers = {"accept": "application/json", "Content-Type": "application/json"}
+        data = {
+            "client_id": "ownerapi",
+            "grant_type": "authorization_code",
+            "code": str(code.group(1)),
+            "code_verifier": self.__apiVerifier.decode("UTF-8"),
+            "redirect_uri": self.__callbackURL,
+        }
+        req = None
+        now = time.time()
+        try:
+            req = requests.post(self.__authURL, headers=headers, json=data)
+            logger.log(logging.INFO2, "Car API request" + str(req))
+            apiResponseDict = json.loads(req.text)
+        except requests.exceptions.RequestException:
+            logger.error("Request Exception parsing API Token Exchange Response")
+            pass
+        except ValueError:
+            pass
+        except json.decoder.JSONDecodeError:
+            logger.error("JSON Decode Error parsing API Token Exchange Response")
+            pass
+
+        params = json.loads(req.text)
+
+        # Check for errors
+        if "error" in params:
+            return params["error"]
+
+        if "access_token" in params:
+            try:
+                self.setCarApiBearerToken(params["access_token"])
+                self.setCarApiRefreshToken(params["refresh_token"])
+                self.setCarApiTokenExpireTime(time.time() + params["expires_in"])
+                self.master.queue_background_task({"cmd": "saveSettings"})
+                return "success"
+            except KeyError:
+                logger.log(
+                    logging.INFO2,
+                    "ERROR: Can't access Tesla car via API.  Please log in again via web interface.",
+                )
+                self.updateCarApiLastErrorTime()
+                return "response_no_token"
+
+        logger.log(logging.INFO2, str(req))
+        logger.log(logging.INFO2, req.text)
+
+        return "unknown"
+
     def setCarApiBearerToken(self, token=None):
         if token:
-            self.carApiBearerToken = token
-            return True
+            if self.master.tokenSyncEnabled():
+                # We won't accept tokens if Token Sync is already in place
+                return False
+            else:
+                self.carApiBearerToken = token
+                return True
         else:
             return False
 
@@ -1237,9 +1090,36 @@ class TeslaAPI:
         self.carApiTokenExpireTime = value
         return True
 
-    def submitCaptchaCode(self, code):
-        self.__apiCaptchaCode = code
-        return self.apiLoginPhaseOne()
+    def setChargeRate(self, charge_rate, vehicle=None):
+
+        # As a fallback to allow initial implementation of the charge rate functionality for single car installs,
+        # If no vehcle is specified, we take the first returned to us.
+
+        if not vehicle:
+            vehicle = self.getCarApiVehicles()[0]
+
+        vehicle.lastAPIAccessTime = time.time()
+
+        url = "https://owner-api.teslamotors.com/api/1/vehicles/"
+        url = url + str(vehicle.ID) + "/command/set_charging_amps"
+
+        headers = {
+            "accept": "application/json",
+            "Authorization": "Bearer " + self.getCarApiBearerToken(),
+        }
+
+        body = {"charging_amps": charge_rate}
+
+        try:
+            req = requests.post(url, headers=headers, json=body)
+            logger.log(logging.INFO8, "Car API cmd set_charging_amps" + str(req))
+            apiResponseDict = json.loads(req.text)
+        except requests.exceptions.RequestException:
+            return False
+        except json.decoder.JSONDecodeError:
+            return False
+
+        return apiResponseDict
 
     def updateCarApiLastErrorTime(self, vehicle=None):
         timestamp = time.time()
@@ -1269,6 +1149,28 @@ class TeslaAPI:
                 car.update_charge()
         self.lastChargeCheck = time.time()
 
+    def wakeVehicle(self, vehicle):
+        apiResponseDict = None
+        vehicle.lastAPIAccessTime = time.time()
+
+        url = "https://owner-api.teslamotors.com/api/1/vehicles/"
+        url = url + str(vehicle.ID) + "/wake_up"
+
+        headers = {
+            "accept": "application/json",
+            "Authorization": "Bearer " + self.getCarApiBearerToken(),
+        }
+        try:
+            req = requests.post(url, headers=headers)
+            logger.log(logging.INFO8, "Car API cmd wake_up" + str(req))
+            apiResponseDict = json.loads(req.text)
+        except requests.exceptions.RequestException:
+            return False
+        except json.decoder.JSONDecodeError:
+            return False
+
+        return apiResponseDict
+
     @property
     def numCarsAtHome(self):
         return len([car for car in self.carApiVehicles if car.atHome])
@@ -1284,15 +1186,13 @@ class TeslaAPI:
 
 
 class CarApiVehicle:
-    import time
-    import requests
-    import json
 
     carapi = None
-    config = None
+    __config = None
     debuglevel = 0
     ID = None
     name = ""
+    syncSource = "TeslaAPI"
     VIN = ""
 
     firstWakeAttemptTime = 0
@@ -1314,13 +1214,49 @@ class CarApiVehicle:
     atHome = False
     timeToFullCharge = 0.0
 
+    # Sync values are updated by an external module such as TeslaMate
+    syncTimestamp = 0
+    syncTimeout = 60 * 60
+    syncLat = 10000
+    syncLon = 10000
+    syncState = "asleep"
+
     def __init__(self, json, carapi, config):
         self.carapi = carapi
-        self.config = config
+        self.__config = config
         self.ID = json["id"]
         self.VIN = json["vin"]
         self.name = json["display_name"]
 
+        # Launch sync monitoring thread
+        Thread(target=self.checkSyncNotStale).start()
+
+    def checkSyncNotStale(self):
+        # Once an external system begins providing sync functionality to defer
+        # Tesla API queries and provide already fetched information, there is a
+        # potential condition which may occur in which the external system goes
+        # away and leaves us with stale data.
+
+        # To guard against this, this threaded function will loop every x minutes
+        # and check the last sync timestamp. If it has not updated in that interval,
+        # we switch back to using the API
+
+        while True:
+            if (
+                self.syncSource != "TeslaAPI"
+                and self.self.is_awake()
+                and (self.syncTimestamp < (time.time() - self.syncTimeout))
+            ):
+                logger.error(
+                    "Data from "
+                    + self.syncSource
+                    + " for "
+                    + self.name
+                    + " is stale. Switching back to TeslaAPI"
+                )
+                self.syncSource = "TeslaAPI"
+            time.sleep(self.syncTimeout)
+
     def ready(self):
         if self.carapi.getCarApiRetryRemaining(self):
             # It's been under carApiErrorRetryMins minutes since the car API
@@ -1356,9 +1292,19 @@ class CarApiVehicle:
 
     # Permits opportunistic API requests
     def is_awake(self):
-        url = "https://owner-api.teslamotors.com/api/1/vehicles/" + str(self.ID)
-        (result, response) = self.get_car_api(url, checkReady=False, provesOnline=False)
-        return result and response.get("state", "") == "online"
+        if self.syncSource == "TeslaAPI":
+            url = "https://owner-api.teslamotors.com/api/1/vehicles/" + str(self.ID)
+            (result, response) = self.get_car_api(
+                url, checkReady=False, provesOnline=False
+            )
+            return result and response.get("state", "") == "online"
+        else:
+            return (
+                self.syncState == "online"
+                or self.syncState == "charging"
+                or self.syncState == "updating"
+                or self.syncState == "driving"
+            )
 
     def get_car_api(self, url, checkReady=True, provesOnline=True):
         if checkReady and not self.ready():
@@ -1423,50 +1369,66 @@ class CarApiVehicle:
 
     def update_location(self, cacheTime=60):
 
-        url = "https://owner-api.teslamotors.com/api/1/vehicles/"
-        url = url + str(self.ID) + "/data_request/drive_state"
+        if self.syncSource == "TeslaAPI":
+            url = "https://owner-api.teslamotors.com/api/1/vehicles/"
+            url = url + str(self.ID) + "/data_request/drive_state"
 
-        now = time.time()
+            now = time.time()
 
-        if now - self.lastDriveStatusTime < cacheTime:
-            return True
+            if now - self.lastDriveStatusTime < cacheTime:
+                return True
 
-        try:
-            (result, response) = self.get_car_api(url)
-        except TypeError:
-            logger.log(logging.error, "Got None response from get_car_api()")
-            return False
+            try:
+                (result, response) = self.get_car_api(url)
+            except TypeError:
+                logger.log(logging.error, "Got None response from get_car_api()")
+                return False
 
-        if result:
-            self.lastDriveStatusTime = now
-            self.lat = response["latitude"]
-            self.lon = response["longitude"]
+            if result:
+                self.lastDriveStatusTime = now
+                self.lat = response["latitude"]
+                self.lon = response["longitude"]
+                self.atHome = self.carapi.is_location_home(self.lat, self.lon)
+
+            return result
+
+        else:
+
+            self.lat = self.syncLat
+            self.lon = self.syncLon
             self.atHome = self.carapi.is_location_home(self.lat, self.lon)
 
-        return result
-
-    def update_charge(self):
-        url = "https://owner-api.teslamotors.com/api/1/vehicles/"
-        url = url + str(self.ID) + "/data_request/charge_state"
-
-        now = time.time()
-
-        if now - self.lastChargeStatusTime < 60:
             return True
 
-        try:
-            (result, response) = self.get_car_api(url)
-        except TypeError:
-            logger.log(logging.error, "Got None response from get_car_api()")
-            return False
+    def update_charge(self):
 
-        if result:
-            self.lastChargeStatusTime = time.time()
-            self.chargeLimit = response["charge_limit_soc"]
-            self.batteryLevel = response["battery_level"]
-            self.timeToFullCharge = response["time_to_full_charge"]
+        if self.syncSource == "TeslaAPI":
 
-        return result
+            url = "https://owner-api.teslamotors.com/api/1/vehicles/"
+            url = url + str(self.ID) + "/data_request/charge_state"
+
+            now = time.time()
+
+            if now - self.lastChargeStatusTime < 60:
+                return True
+
+            try:
+                (result, response) = self.get_car_api(url)
+            except TypeError:
+                logger.log(logging.error, "Got None response from get_car_api()")
+                return False
+
+            if result:
+                self.lastChargeStatusTime = time.time()
+                self.chargeLimit = response["charge_limit_soc"]
+                self.batteryLevel = response["battery_level"]
+                self.timeToFullCharge = response["time_to_full_charge"]
+
+            return result
+
+        else:
+
+            return True
 
     def apply_charge_limit(self, limit):
         if self.stopTryingToApplyLimit:
